@@ -1,15 +1,16 @@
+from http.server import HTTPServer, SimpleHTTPRequestHandler
 import logging
 import queue
+from socketserver import ThreadingMixIn
 import threading
+import time
 import traceback
 
-from http.server import SimpleHTTPRequestHandler, HTTPServer
-from socketserver import ThreadingMixIn
-
 import ray.cloudpickle as pickle
-from ray.rllib.offline.input_reader import InputReader
 from ray.rllib.env.policy_client import PolicyClient, \
     _create_embedded_rollout_worker
+from ray.rllib.offline.input_reader import InputReader
+from ray.rllib.policy.sample_batch import SampleBatch
 from ray.rllib.utils.annotations import override, PublicAPI
 
 logger = logging.getLogger(__name__)
@@ -45,7 +46,7 @@ class PolicyServerInput(ThreadingMixIn, HTTPServer, InputReader):
     """
 
     @PublicAPI
-    def __init__(self, ioctx, address, port):
+    def __init__(self, ioctx, address, port, idle_timeout=3.0):
         """Create a PolicyServerInput.
 
         This class implements rllib.offline.InputReader, and can be used with
@@ -67,6 +68,7 @@ class PolicyServerInput(ThreadingMixIn, HTTPServer, InputReader):
         self.rollout_worker = ioctx.worker
         self.samples_queue = queue.Queue()
         self.metrics_queue = queue.Queue()
+        self.idle_timeout = idle_timeout
 
         def get_metrics():
             completed = []
@@ -80,22 +82,46 @@ class PolicyServerInput(ThreadingMixIn, HTTPServer, InputReader):
         # Forwards client-reported rewards directly into the local rollout
         # worker. This is a bit of a hack since it is patching the get_metrics
         # function of the sampler.
-        self.rollout_worker.sampler.get_metrics = get_metrics
+        if self.rollout_worker.sampler is not None:
+            self.rollout_worker.sampler.get_metrics = get_metrics
 
+        # Create a request handler that receives commands from the clients
+        # and sends data and metrics into the queues.
         handler = _make_handler(self.rollout_worker, self.samples_queue,
                                 self.metrics_queue)
         HTTPServer.__init__(self, (address, port), handler)
-        logger.info("")
+
         logger.info("Starting connector server at {}:{}".format(address, port))
-        logger.info("")
-        thread = threading.Thread(name="server", target=self.serve_forever)
-        thread.daemon = True
-        thread.start()
+
+        # Start the serving thread, listening on socket and handling commands.
+        serving_thread = threading.Thread(
+            name="server", target=self.serve_forever)
+        serving_thread.daemon = True
+        serving_thread.start()
+
+        # Start a dummy thread that puts empty SampleBatches on the queue, just
+        # in case we don't receive anything from clients (or there aren't
+        # any). The latter would block sample collection entirely otherwise,
+        # even if other workers' PolicyServerInput receive incoming data from
+        # actual clients.
+        heart_beat_thread = threading.Thread(
+            name="heart-beat", target=self._put_empty_sample_batch_every_n_sec)
+        heart_beat_thread.daemon = True
+        heart_beat_thread.start()
 
     @override(InputReader)
     def next(self):
         return self.samples_queue.get()
 
+    def _put_empty_sample_batch_every_n_sec(self):
+        # Places an empty SampleBatch every `idle_timeout` seconds onto the
+        # `samples_queue`. This avoids hanging of all RolloutWorkers parallel
+        # to this one in case this PolicyServerInput does not have incoming
+        # data (e.g. no client connected).
+        while True:
+            time.sleep(self.idle_timeout)
+            self.samples_queue.put(SampleBatch())
+
 
 def _make_handler(rollout_worker, samples_queue, metrics_queue):
     # Only used in remote inference mode. We must create a new rollout worker
@@ -149,6 +175,7 @@ def _make_handler(rollout_worker, samples_queue, metrics_queue):
         def execute_command(self, args):
             command = args["command"]
             response = {}
+
             # Local inference commands:
             if command == PolicyClient.GET_WORKER_ARGS:
                 logger.info("Sending worker creation args to client.")
@@ -161,6 +188,7 @@ def _make_handler(rollout_worker, samples_queue, metrics_queue):
                 logger.info("Got sample batch of size {} from client.".format(
                     args["samples"].count))
                 report_data(args)
+
             # Remote inference commands:
             elif command == PolicyClient.START_EPISODE:
                 setup_child_rollout_worker()
