@@ -14,11 +14,6 @@ if __name__ == "__main__":
     env = env_creator({})
     register_env("waterworld", env_creator)
 
-    obs_space = env.observation_space
-    act_spc = env.action_space
-
-    policies = {agent: (None, obs_space, act_spc, {}) for agent in env.agents}
-
     tune.run(
         "APEX_DDPG",
         stop={"episodes_total": 60000},
@@ -31,8 +26,9 @@ if __name__ == "__main__":
             "num_workers": 2,
             # Method specific
             "multiagent": {
-                "policies": policies,
-                "policy_mapping_fn": (lambda agent_id: agent_id),
+                "policies": set(env.agents),
+                "policy_mapping_fn": (
+                    lambda agent_id, episode, **kwargs: agent_id),
             },
         },
     )
