@@ -1,12 +1,11 @@
 import logging
 import numpy as np
 import math
-import tree
+import tree  # pip install dm_tree
 from typing import List, Tuple, Any
 
 import ray
-from ray.rllib.evaluation.metrics import extract_stats, get_learner_stats, \
-    LEARNER_STATS_KEY
+from ray.rllib.evaluation.metrics import get_learner_stats
 from ray.rllib.evaluation.worker_set import WorkerSet
 from ray.rllib.execution.common import \
     AGENT_STEPS_TRAINED_COUNTER, APPLY_GRADS_TIMER, COMPUTE_GRADS_TIMER, \
@@ -14,7 +13,7 @@ from ray.rllib.execution.common import \
     LOAD_BATCH_TIMER, NUM_TARGET_UPDATES, STEPS_SAMPLED_COUNTER, \
     STEPS_TRAINED_COUNTER, WORKER_UPDATE_TIMER, _check_sample_batch_type, \
     _get_global_vars, _get_shared_metrics
-from ray.rllib.execution.multi_gpu_impl import LocalSyncParallelOptimizer
+from ray.rllib.policy.policy import LEARNER_STATS_KEY
 from ray.rllib.policy.sample_batch import SampleBatch, DEFAULT_POLICY_ID, \
     MultiAgentBatch
 from ray.rllib.utils.framework import try_import_tf
@@ -48,7 +47,8 @@ class TrainOneStep:
                  num_sgd_iter: int = 1,
                  sgd_minibatch_size: int = 0):
         self.workers = workers
-        self.policies = policies or workers.local_worker().policies_to_train
+        self.local_worker = workers.local_worker()
+        self.policies = policies
         self.num_sgd_iter = num_sgd_iter
         self.sgd_minibatch_size = sgd_minibatch_size
 
@@ -58,22 +58,21 @@ class TrainOneStep:
         metrics = _get_shared_metrics()
         learn_timer = metrics.timers[LEARN_ON_BATCH_TIMER]
         with learn_timer:
+            # Subsample minibatches (size=`sgd_minibatch_size`) from the
+            # train batch and loop through train batch `num_sgd_iter` times.
             if self.num_sgd_iter > 1 or self.sgd_minibatch_size > 0:
                 lw = self.workers.local_worker()
                 info = do_minibatch_sgd(
-                    batch, {pid: lw.get_policy(pid)
-                            for pid in self.policies}, lw, self.num_sgd_iter,
-                    self.sgd_minibatch_size, [])
-                # TODO(ekl) shouldn't be returning learner stats directly here
-                # TODO(sven): Skips `custom_metrics` key from on_learn_on_batch
-                #  callback (shouldn't).
-                metrics.info[LEARNER_INFO] = info
+                    batch, {
+                        pid: lw.get_policy(pid)
+                        for pid in self.policies
+                        or self.local_worker.policies_to_train
+                    }, lw, self.num_sgd_iter, self.sgd_minibatch_size, [])
+            # Single update step using train batch.
             else:
                 info = self.workers.local_worker().learn_on_batch(batch)
-                metrics.info[LEARNER_INFO] = extract_stats(
-                    info, LEARNER_STATS_KEY)
-                metrics.info["custom_metrics"] = extract_stats(
-                    info, "custom_metrics")
+
+            metrics.info[LEARNER_INFO] = info
             learn_timer.push_units_processed(batch.count)
         metrics.counters[STEPS_TRAINED_COUNTER] += batch.count
         if isinstance(batch, MultiAgentBatch):
@@ -84,7 +83,7 @@ class TrainOneStep:
         if self.workers.remote_workers():
             with metrics.timers[WORKER_UPDATE_TIMER]:
                 weights = ray.put(self.workers.local_worker().get_weights(
-                    self.policies))
+                    self.policies or self.local_worker.policies_to_train))
                 for e in self.workers.remote_workers():
                     e.set_weights.remote(weights, _get_global_vars())
         # Also update global vars of the local worker.
@@ -92,15 +91,15 @@ class TrainOneStep:
         return batch, info
 
 
-class TrainTFMultiGPU:
-    """TF Multi-GPU version of TrainOneStep.
+class MultiGPUTrainOneStep:
+    """Multi-GPU version of TrainOneStep.
 
     This should be used with the .for_each() operator. A tuple of the input
     and learner stats will be returned.
 
     Examples:
         >>> rollouts = ParallelRollouts(...)
-        >>> train_op = rollouts.for_each(TrainMultiGPU(workers, ...))
+        >>> train_op = rollouts.for_each(MultiGPUTrainOneStep(workers, ...))
         >>> print(next(train_op))  # This trains the policy on one batch.
         SampleBatch(...), {"learner_stats": ...}
 
@@ -115,11 +114,10 @@ class TrainTFMultiGPU:
                  num_sgd_iter: int,
                  num_gpus: int,
                  shuffle_sequences: bool,
-                 policies: List[PolicyID] = frozenset([]),
                  _fake_gpus: bool = False,
                  framework: str = "tf"):
         self.workers = workers
-        self.policies = policies or workers.local_worker().policies_to_train
+        self.local_worker = workers.local_worker()
         self.num_sgd_iter = num_sgd_iter
         self.sgd_minibatch_size = sgd_minibatch_size
         self.shuffle_sequences = shuffle_sequences
@@ -131,48 +129,23 @@ class TrainTFMultiGPU:
             num_gpus = 1
         type_ = "cpu" if _fake_gpus else "gpu"
         self.devices = [
-            "/{}:{}".format(type_, i) for i in range(int(math.ceil(num_gpus)))
+            "/{}:{}".format(type_, 0 if _fake_gpus else i)
+            for i in range(int(math.ceil(num_gpus)))
         ]
 
-        # Total batch size (all towers). Make sure it is dividable by
-        # num towers.
-        self.batch_size = int(sgd_minibatch_size / len(self.devices)) * len(
-            self.devices)
-        assert self.batch_size % len(self.devices) == 0
-        assert self.batch_size >= len(self.devices), "batch size too small"
+        # Make sure total batch size is dividable by the number of devices.
         # Batch size per tower.
-        self.per_device_batch_size = int(self.batch_size / len(self.devices))
-
-        # per-GPU graph copies created below must share vars with the policy
-        # reuse is set to AUTO_REUSE because Adam nodes are created after
-        # all of the device copies are created.
-        self.optimizers = {}
-        with self.workers.local_worker().tf_sess.graph.as_default():
-            with self.workers.local_worker().tf_sess.as_default():
-                for policy_id in self.policies:
-                    policy = self.workers.local_worker().get_policy(policy_id)
-                    with tf1.variable_scope(policy_id, reuse=tf1.AUTO_REUSE):
-                        if policy._state_inputs:
-                            rnn_inputs = policy._state_inputs + [
-                                policy._seq_lens
-                            ]
-                        else:
-                            rnn_inputs = []
-                        self.optimizers[policy_id] = (
-                            LocalSyncParallelOptimizer(
-                                policy._optimizer, self.devices,
-                                list(policy._loss_input_dict_no_rnn.values()),
-                                rnn_inputs, self.per_device_batch_size,
-                                policy.copy))
-
-                self.sess = self.workers.local_worker().tf_sess
-                self.sess.run(tf1.global_variables_initializer())
+        self.per_device_batch_size = sgd_minibatch_size // len(self.devices)
+        # Total batch size.
+        self.batch_size = self.per_device_batch_size * len(self.devices)
+        assert self.batch_size % len(self.devices) == 0
+        assert self.batch_size >= len(self.devices), "Batch size too small!"
 
     def __call__(self,
                  samples: SampleBatchType) -> (SampleBatchType, List[dict]):
         _check_sample_batch_type(samples)
 
-        # Handle everything as if multiagent
+        # Handle everything as if multi agent.
         if isinstance(samples, SampleBatch):
             samples = MultiAgentBatch({
                 DEFAULT_POLICY_ID: samples
@@ -183,56 +156,66 @@ class TrainTFMultiGPU:
         learn_timer = metrics.timers[LEARN_ON_BATCH_TIMER]
         # Load data into GPUs.
         with load_timer:
-            num_loaded_tuples = {}
+            num_loaded_samples = {}
             for policy_id, batch in samples.policy_batches.items():
                 # Not a policy-to-train.
-                if policy_id not in self.policies:
+                if policy_id not in self.local_worker.policies_to_train:
                     continue
 
                 # Decompress SampleBatch, in case some columns are compressed.
                 batch.decompress_if_needed()
 
-                policy = self.workers.local_worker().get_policy(policy_id)
-                policy._debug_vars()
-                tuples = policy._get_loss_inputs_dict(
-                    batch, shuffle=self.shuffle_sequences)
-                data_keys = list(policy._loss_input_dict_no_rnn.values())
-                if policy._state_inputs:
-                    state_keys = policy._state_inputs + [policy._seq_lens]
-                else:
-                    state_keys = []
-                num_loaded_tuples[policy_id] = (
-                    self.optimizers[policy_id].load_data(
-                        self.sess, [tuples[k] for k in data_keys],
-                        [tuples[k] for k in state_keys]))
+                # Load the entire train batch into the Policy's only buffer
+                # (idx=0). Policies only have >1 buffers, if we are training
+                # asynchronously.
+                num_loaded_samples[policy_id] = self.local_worker.policy_map[
+                    policy_id].load_batch_into_buffer(
+                        batch, buffer_index=0)
 
         # Execute minibatch SGD on loaded data.
         with learn_timer:
             fetches = {}
-            for policy_id, tuples_per_device in num_loaded_tuples.items():
-                optimizer = self.optimizers[policy_id]
+            for policy_id, samples_per_device in num_loaded_samples.items():
+                policy = self.local_worker.policy_map[policy_id]
                 num_batches = max(
                     1,
-                    int(tuples_per_device) // int(self.per_device_batch_size))
+                    int(samples_per_device) // int(self.per_device_batch_size))
                 logger.debug("== sgd epochs for {} ==".format(policy_id))
+                batch_fetches_all_towers = []
                 for _ in range(self.num_sgd_iter):
                     permutation = np.random.permutation(num_batches)
-                    batch_fetches_all_towers = []
                     for batch_index in range(num_batches):
-                        batch_fetches = optimizer.optimize(
-                            self.sess, permutation[batch_index] *
-                            self.per_device_batch_size)
+                        # Learn on the pre-loaded data in the buffer.
+                        # Note: For minibatch SGD, the data is an offset into
+                        # the pre-loaded entire train batch.
+                        batch_fetches = policy.learn_on_loaded_batch(
+                            permutation[batch_index] *
+                            self.per_device_batch_size,
+                            buffer_index=0)
 
-                        batch_fetches_all_towers.append(
-                            tree.map_structure_with_path(
-                                lambda p, *s: self._all_tower_reduce(p, *s),
-                                *(batch_fetches["tower_{}".format(tower_num)]
-                                  for tower_num in range(len(self.devices)))))
+                        # No towers: Single CPU.
+                        if "tower_0" not in batch_fetches:
+                            batch_fetches_all_towers.append(batch_fetches)
+                        else:
+                            batch_fetches_all_towers.append(
+                                tree.map_structure_with_path(
+                                    lambda p, *s: all_tower_reduce(p, *s),
+                                    *(batch_fetches.pop(
+                                        "tower_{}".format(tower_num))
+                                      for tower_num in range(
+                                          len(self.devices)))))
+                            for k, v in batch_fetches.items():
+                                if k == LEARNER_STATS_KEY:
+                                    for k1, v1 in batch_fetches[k].items():
+                                        batch_fetches_all_towers[-1][
+                                            LEARNER_STATS_KEY][k1] = v1
+                                else:
+                                    batch_fetches_all_towers[-1][k] = v
 
                 # Reduce mean across all minibatch SGD steps (axis=0 to keep
                 # all shapes as-is).
                 fetches[policy_id] = tree.map_structure(
-                    lambda *s: np.nanmean(s, axis=0),
+                    lambda *s: None if s[0] is None else np.nanmean(s, axis=0),
                     *batch_fetches_all_towers)
 
         load_timer.push_units_processed(samples.count)
@@ -241,25 +224,32 @@ class TrainTFMultiGPU:
         metrics.counters[STEPS_TRAINED_COUNTER] += samples.count
         metrics.counters[AGENT_STEPS_TRAINED_COUNTER] += samples.agent_steps()
         metrics.info[LEARNER_INFO] = fetches
+
         if self.workers.remote_workers():
             with metrics.timers[WORKER_UPDATE_TIMER]:
                 weights = ray.put(self.workers.local_worker().get_weights(
-                    self.policies))
+                    self.local_worker.policies_to_train))
                 for e in self.workers.remote_workers():
                     e.set_weights.remote(weights, _get_global_vars())
+
         # Also update global vars of the local worker.
         self.workers.local_worker().set_global_vars(_get_global_vars())
         return samples, fetches
 
-    def _all_tower_reduce(self, path, *tower_data):
-        """Reduces stats across towers based on their stats-dict paths."""
-        if len(path) == 1 and path[0] == "td_error":
-            return np.concatenate(tower_data, axis=0)
-        elif path[-1].startswith("min_"):
-            return np.nanmin(tower_data)
-        elif path[-1].startswith("max_"):
-            return np.nanmax(tower_data)
-        return np.nanmean(tower_data)
+
+# Backward compatibility.
+TrainTFMultiGPU = MultiGPUTrainOneStep
+
+
+def all_tower_reduce(path, *tower_data):
+    """Reduces stats across towers based on their stats-dict paths."""
+    if len(path) == 1 and path[0] == "td_error":
+        return np.concatenate(tower_data, axis=0)
+    elif path[-1].startswith("min_"):
+        return np.nanmin(tower_data)
+    elif path[-1].startswith("max_"):
+        return np.nanmax(tower_data)
+    return np.nanmean(tower_data)
 
 
 class ComputeGradients:
@@ -313,7 +303,8 @@ class ApplyGradients:
                 currently processing (i.e., A3C style).
         """
         self.workers = workers
-        self.policies = policies or workers.local_worker().policies_to_train
+        self.local_worker = workers.local_worker()
+        self.policies = policies
         self.update_all = update_all
 
     def __call__(self, item: Tuple[ModelGradients, int]) -> None:
@@ -337,7 +328,7 @@ class ApplyGradients:
             if self.workers.remote_workers():
                 with metrics.timers[WORKER_UPDATE_TIMER]:
                     weights = ray.put(self.workers.local_worker().get_weights(
-                        self.policies))
+                        self.policies or self.local_worker.policies_to_train))
                     for e in self.workers.remote_workers():
                         e.set_weights.remote(weights, _get_global_vars())
         else:
@@ -348,7 +339,7 @@ class ApplyGradients:
                     "in the iterator context.")
             with metrics.timers[WORKER_UPDATE_TIMER]:
                 weights = self.workers.local_worker().get_weights(
-                    self.policies)
+                    self.policies or self.local_worker.policies_to_train)
                 metrics.current_actor.set_weights.remote(
                     weights, _get_global_vars())
 
@@ -405,8 +396,9 @@ class UpdateTargetNetwork:
                  by_steps_trained: bool = False,
                  policies: List[PolicyID] = frozenset([])):
         self.workers = workers
+        self.local_worker = workers.local_worker()
         self.target_update_freq = target_update_freq
-        self.policies = (policies or workers.local_worker().policies_to_train)
+        self.policies = policies
         if by_steps_trained:
             self.metric = STEPS_TRAINED_COUNTER
         else:
@@ -417,7 +409,7 @@ class UpdateTargetNetwork:
         cur_ts = metrics.counters[self.metric]
         last_update = metrics.counters[LAST_TARGET_UPDATE_TS]
         if cur_ts - last_update > self.target_update_freq:
-            to_update = self.policies
+            to_update = self.policies or self.local_worker.policies_to_train
             self.workers.local_worker().foreach_trainable_policy(
                 lambda p, p_id: p_id in to_update and p.update_target())
             metrics.counters[NUM_TARGET_UPDATES] += 1
