@@ -8,7 +8,6 @@ import unittest
 import ray
 from ray.rllib.agents.a3c import A2CTrainer
 from ray.rllib.agents.pg import PGTrainer
-from ray.rllib.agents.pg.pg_tf_policy import PGTFPolicy
 from ray.rllib.env import MultiAgentEnv
 from ray.rllib.env.base_env import BaseEnv
 from ray.rllib.env.vector_env import VectorEnv
@@ -451,10 +450,10 @@ class NestedSpacesTest(unittest.TestCase):
                 "multiagent": {
                     "policies": {
                         "tuple_policy": (
-                            PGTFPolicy, TUPLE_SPACE, act_space,
+                            None, TUPLE_SPACE, act_space,
                             {"model": {"custom_model": "tuple_spy"}}),
                         "dict_policy": (
-                            PGTFPolicy, DICT_SPACE, act_space,
+                            None, DICT_SPACE, act_space,
                             {"model": {"custom_model": "dict_spy"}}),
                     },
                     "policy_mapping_fn": lambda aid, **kwargs: {
