@@ -131,7 +131,7 @@ class RayServeReplica:
 
     def __init__(self, _callable: Callable, backend_config: BackendConfig,
                  is_function: bool, controller_handle: ActorHandle) -> None:
-        self.backend_tag = ray.serve.api.get_replica_context().backend_tag
+        self.backend_tag = ray.serve.api.get_replica_context().deployment
         self.replica_tag = ray.serve.api.get_replica_context().replica_tag
         self.callable = _callable
         self.is_function = is_function
@@ -141,11 +141,11 @@ class RayServeReplica:
         self.num_ongoing_requests = 0
 
         self.request_counter = metrics.Counter(
-            "serve_backend_request_counter",
+            "serve_deployment_request_counter",
             description=("The number of queries that have been "
                          "processed in this replica."),
-            tag_keys=("backend", ))
-        self.request_counter.set_default_tags({"backend": self.backend_tag})
+            tag_keys=("deployment", ))
+        self.request_counter.set_default_tags({"deployment": self.backend_tag})
 
         self.loop = asyncio.get_event_loop()
         self.long_poll_client = LongPollClient(
@@ -158,38 +158,38 @@ class RayServeReplica:
         )
 
         self.error_counter = metrics.Counter(
-            "serve_backend_error_counter",
+            "serve_deployment_error_counter",
             description=("The number of exceptions that have "
-                         "occurred in the backend."),
-            tag_keys=("backend", ))
-        self.error_counter.set_default_tags({"backend": self.backend_tag})
+                         "occurred in the deployment."),
+            tag_keys=("deployment", ))
+        self.error_counter.set_default_tags({"deployment": self.backend_tag})
 
         self.restart_counter = metrics.Counter(
-            "serve_backend_replica_starts",
+            "serve_deployment_replica_starts",
             description=("The number of times this replica "
                          "has been restarted due to failure."),
-            tag_keys=("backend", "replica"))
+            tag_keys=("deployment", "replica"))
         self.restart_counter.set_default_tags({
-            "backend": self.backend_tag,
+            "deployment": self.backend_tag,
             "replica": self.replica_tag
         })
 
         self.processing_latency_tracker = metrics.Histogram(
-            "serve_backend_processing_latency_ms",
+            "serve_deployment_processing_latency_ms",
             description="The latency for queries to be processed.",
             boundaries=DEFAULT_LATENCY_BUCKET_MS,
-            tag_keys=("backend", "replica"))
+            tag_keys=("deployment", "replica"))
         self.processing_latency_tracker.set_default_tags({
-            "backend": self.backend_tag,
+            "deployment": self.backend_tag,
             "replica": self.replica_tag
         })
 
         self.num_processing_items = metrics.Gauge(
             "serve_replica_processing_queries",
             description="The current number of queries being processed.",
-            tag_keys=("backend", "replica"))
+            tag_keys=("deployment", "replica"))
         self.num_processing_items.set_default_tags({
-            "backend": self.backend_tag,
+            "deployment": self.backend_tag,
             "replica": self.replica_tag
         })
 
@@ -200,7 +200,7 @@ class RayServeReplica:
             handler.setFormatter(
                 logging.Formatter(
                     handler.formatter._fmt +
-                    f" component=serve backend={self.backend_tag} "
+                    f" component=serve deployment={self.backend_tag} "
                     f"replica={self.replica_tag}"))
 
     def get_runner_method(self, request_item: Query) -> Callable:
